# Copyright (C) 2014 Ivan Melnikov <iv at altlinux dot org>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import errno
import os
import threading

from docutils import frontend
from docutils import parsers as docutils_parser
from docutils import utils
import restructuredtext_lint as rl


class ParsedFile(object):
    FALLBACK_ENCODING = "utf-8"

    def __init__(self, filename, encoding=None, default_extension=""):
        self._filename = filename
        self._content = None
        self._raw_content = None
        self._encoding = encoding
        self._doc = None
        self._errors = None
        self._lines = None
        self._has_read = False
        self._extension = os.path.splitext(filename)[1]
        self._read_lock = threading.Lock()
        if not self._extension:
            self._extension = default_extension

    @property
    def errors(self):
        if self._errors is not None:
            return self._errors
        self._errors = rl.lint(self.contents, filepath=self.filename)
        return self._errors

    @property
    def document(self):
        if self._doc is None:
            # Use the rst parsers document output to do as much of the
            # validation as we can without resorting to custom logic (this
            # parser is what sphinx and others use anyway so it's hopefully
            # mature).
            parser_cls = docutils_parser.get_parser_class("rst")
            parser = parser_cls()
            defaults = {
                "halt_level": 5,
                "report_level": 5,
                "quiet": True,
                "file_insertion_enabled": False,
                "traceback": True,
                # Development use only.
                "dump_settings": False,
                "dump_internals": False,
                "dump_transforms": False,
            }
            opt = frontend.OptionParser(components=[parser], defaults=defaults)
            doc = utils.new_document(
                source_path=self.filename, settings=opt.get_default_values()
            )
            parser.parse(self.contents, doc)
            self._doc = doc
        return self._doc

    def _read(self):
        if self._has_read:
            return
        with self._read_lock:
            if not self._has_read:
                with open(self.filename, "rb") as fh:
                    self._raw_content = fh.read()
                    self._lines = self._raw_content.splitlines(True)
                self._has_read = True

    def lines_iter(self, remove_trailing_newline=True):
        self._read()
        for line in self._lines:
            line = str(line, encoding=self.encoding)
            if remove_trailing_newline:
                # Cope with various OS new line conventions
                if line.endswith("\n"):
                    line = line[:-1]
                if line.endswith("\r"):
                    line = line[:-1]
            yield line

    @property
    def lines(self):
        self._read()
        return self._lines

    @property
    def extension(self):
        return self._extension

    @property
    def filename(self):
        return self._filename

    @property
    def encoding(self):
        if not self._encoding:
            self._encoding = self.FALLBACK_ENCODING
        return self._encoding

    @property
    def raw_contents(self):
        self._read()
        return self._raw_content

    @property
    def contents(self):
        if self._content is None:
            self._content = str(self.raw_contents, encoding=self.encoding)
        return self._content

    def __str__(self):
        return "%s (%s, %s chars, %s lines)" % (
            self.filename,
            self.encoding,
            len(self.contents),
            len(list(self.lines_iter())),
        )


def parse(filename, encoding=None, default_extension=""):
    if not os.path.isfile(filename):
        raise IOError(errno.ENOENT, "File not found", filename)
    return ParsedFile(filename, encoding=encoding, default_extension=default_extension)



# Copyright (C) 2014 Ivan Melnikov <iv at altlinux dot org>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import os


def find_files(paths, extensions, ignored_paths):
    extensions = set(extensions)
    ignored_absolute_paths = set()
    for path in ignored_paths:
        for expanded_path in glob.iglob(path):
            expanded_path = os.path.abspath(expanded_path)
            ignored_absolute_paths.add(expanded_path)

    def extension_matches(path):
        _base, ext = os.path.splitext(path)
        return ext in extensions

    def path_ignorable(path):
        path = os.path.abspath(path)
        if path in ignored_absolute_paths:
            return True
        last_path = None
        while path != last_path:
            # If we hit the root, this loop will stop since the resolution
            # of "/../" is still "/" when ran through the abspath function...
            last_path = path
            path = os.path.abspath(os.path.join(path, os.path.pardir))
            if path in ignored_absolute_paths:
                return True
        return False

    for path in paths:
        if os.path.isfile(path):
            if extension_matches(path):
                yield (path, path_ignorable(path))
        elif os.path.isdir(path):
            for root, dirnames, filenames in os.walk(path):
                for filename in filenames:
                    path = os.path.join(root, filename)
                    if extension_matches(path):
                        yield (path, path_ignorable(path))
        else:
            raise IOError("Invalid path: %s" % path)


def filtered_traverse(document, filter_func):
    for n in document.traverse(include_self=True):
        if filter_func(n):
            yield n


def contains_url(line):
    return "http://" in line or "https://" in line


def has_any_node_type(node, node_types):
    n = node
    while n is not None:
        if isinstance(n, node_types):
            return True
        n = n.parent
    return False



# Copyright (C) 2014 Ivan Melnikov <iv at altlinux dot org>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Check documentation for simple style requirements.

What is checked:
    - invalid rst format - D000
    - lines should not be longer than 79 characters - D001
      - RST exception: line with no whitespace except in the beginning
      - RST exception: lines with http or https urls
      - RST exception: literal blocks
      - RST exception: rst target directives
    - no trailing whitespace - D002
    - no tabulation for indentation - D003
    - no carriage returns (use unix newlines) - D004
    - no newline at end of file - D005
"""

import argparse
import collections
import configparser
import logging
import os
import sys

try:
    import toml

    HAVE_TOML = True
except ImportError:
    HAVE_TOML = False

from stevedore import extension

from doc8 import checks
from doc8 import parser as file_parser
from doc8 import utils
from doc8 import version

FILE_PATTERNS = [".rst", ".txt"]
MAX_LINE_LENGTH = 79
CONFIG_FILENAMES = ["doc8.ini", "tox.ini", "pep8.ini", "setup.cfg"]
if HAVE_TOML:
    CONFIG_FILENAMES.extend(["pyproject.toml"])


def split_set_type(text, delimiter=","):
    return set([i.strip() for i in text.split(delimiter) if i.strip()])


def merge_sets(sets):
    m = set()
    for s in sets:
        m.update(s)
    return m


def parse_ignore_path_errors(entries):
    ignore_path_errors = collections.defaultdict(set)
    for path in entries:
        path, ignored_errors = path.split(";", 1)
        path = path.strip()
        ignored_errors = split_set_type(ignored_errors, delimiter=";")
        ignore_path_errors[path].update(ignored_errors)
    return dict(ignore_path_errors)


def from_ini(fp):
    parser = configparser.RawConfigParser()
    with open(fp, "r") as fh:
        parser.read_file(fh)

    cfg = {}
    try:
        cfg["max_line_length"] = parser.getint("doc8", "max-line-length")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["ignore"] = split_set_type(parser.get("doc8", "ignore"))
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["ignore_path"] = split_set_type(parser.get("doc8", "ignore-path"))
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        ignore_path_errors = parser.get("doc8", "ignore-path-errors")
        ignore_path_errors = split_set_type(ignore_path_errors)
        ignore_path_errors = parse_ignore_path_errors(ignore_path_errors)
        cfg["ignore_path_errors"] = ignore_path_errors
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["allow_long_titles"] = parser.getboolean("doc8", "allow-long-titles")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["sphinx"] = parser.getboolean("doc8", "sphinx")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["verbose"] = parser.getboolean("doc8", "verbose")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["file_encoding"] = parser.get("doc8", "file-encoding")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        cfg["default_extension"] = parser.get("doc8", "default-extension")
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    try:
        extensions = parser.get("doc8", "extensions")
        extensions = extensions.split(",")
        extensions = [s.strip() for s in extensions if s.strip()]
        if extensions:
            cfg["extension"] = extensions
    except (configparser.NoSectionError, configparser.NoOptionError):
        pass
    return cfg


def from_toml(fp):
    cfg = toml.load(fp).get("tool", {}).get("doc8", {})
    return cfg


def extract_config(args):
    cfg = {}
    for cfg_file in args["config"] or CONFIG_FILENAMES:
        if not os.path.isfile(cfg_file):
            if args["config"]:
                print(
                    "Configuration file %s does not exist...ignoring" % (args["config"])
                )
            continue
        if cfg_file.endswith((".ini", ".cfg")):
            cfg = from_ini(cfg_file)
        elif cfg_file.endswith(".toml") and HAVE_TOML:
            cfg = from_toml(cfg_file)
        if cfg:
            break
    return cfg


def fetch_checks(cfg):
    base = [
        checks.CheckValidity(cfg),
        checks.CheckTrailingWhitespace(cfg),
        checks.CheckIndentationNoTab(cfg),
        checks.CheckCarriageReturn(cfg),
        checks.CheckMaxLineLength(cfg),
        checks.CheckNewlineEndOfFile(cfg),
    ]
    mgr = extension.ExtensionManager(
        namespace="doc8.extension.check", invoke_on_load=True, invoke_args=(cfg.copy(),)
    )
    addons = []
    for e in mgr:
        addons.append(e.obj)
    return base + addons


def setup_logging(verbose):
    if verbose:
        level = logging.DEBUG
    else:
        level = logging.ERROR
    logging.basicConfig(
        level=level, format="%(levelname)s: %(message)s", stream=sys.stdout
    )


def scan(cfg):
    if not cfg.get("quiet"):
        print("Scanning...")
    files = collections.deque()
    ignored_paths = cfg.get("ignore_path", [])
    files_ignored = 0
    file_iter = utils.find_files(
        cfg.get("paths", []), cfg.get("extension", []), ignored_paths
    )
    default_extension = cfg.get("default_extension")
    file_encoding = cfg.get("file_encoding")
    for filename, ignoreable in file_iter:
        if ignoreable:
            files_ignored += 1
            if cfg.get("verbose"):
                print("  Ignoring '%s'" % (filename))
        else:
            f = file_parser.parse(
                filename, default_extension=default_extension, encoding=file_encoding
            )
            files.append(f)
            if cfg.get("verbose"):
                print("  Selecting '%s'" % (filename))
    return (files, files_ignored)


def validate(cfg, files, result=None):
    if not cfg.get("quiet"):
        print("Validating...")
    error_counts = {}
    ignoreables = frozenset(cfg.get("ignore", []))
    ignore_targeted = cfg.get("ignore_path_errors", {})
    while files:
        f = files.popleft()
        if cfg.get("verbose"):
            print("Validating %s" % f)
        targeted_ignoreables = set(ignore_targeted.get(f.filename, set()))
        targeted_ignoreables.update(ignoreables)
        for c in fetch_checks(cfg):
            check_name = ".".join([c.__class__.__module__, c.__class__.__name__])
            error_counts.setdefault(check_name, 0)
            try:
                extension_matcher = c.EXT_MATCHER
            except AttributeError:
                pass
            else:
                if not extension_matcher.match(f.extension):
                    if cfg.get("verbose"):
                        print(
                            "  Skipping check '%s' since it does not"
                            " understand parsing a file with extension '%s'"
                            % (check_name, f.extension)
                        )
                    continue
            try:
                reports = set(c.REPORTS)
            except AttributeError:
                pass
            else:
                reports = reports - targeted_ignoreables
                if not reports:
                    if cfg.get("verbose"):
                        print(
                            "  Skipping check '%s', determined to only"
                            " check ignoreable codes" % check_name
                        )
                    continue
            if cfg.get("verbose"):
                print("  Running check '%s'" % check_name)
            if isinstance(c, checks.ContentCheck):
                for line_num, code, message in c.report_iter(f):
                    if code in targeted_ignoreables:
                        continue
                    if not isinstance(line_num, (float, int)):
                        line_num = "?"
                    if cfg.get("verbose"):
                        print(
                            "    - %s:%s: %s %s" % (f.filename, line_num, code, message)
                        )
                    elif not result.capture:
                        print("%s:%s: %s %s" % (f.filename, line_num, code, message))
                    result.error(check_name, f.filename, line_num, code, message)
                    error_counts[check_name] += 1
            elif isinstance(c, checks.LineCheck):
                for line_num, line in enumerate(f.lines_iter(), 1):
                    for code, message in c.report_iter(line):
                        if code in targeted_ignoreables:
                            continue
                        if cfg.get("verbose"):
                            print(
                                "    - %s:%s: %s %s"
                                % (f.filename, line_num, code, message)
                            )
                        elif not result.capture:
                            print(
                                "%s:%s: %s %s" % (f.filename, line_num, code, message)
                            )
                        result.error(check_name, f.filename, line_num, code, message)
                        error_counts[check_name] += 1
            else:
                raise TypeError("Unknown check type: %s, %s" % (type(c), c))
    return error_counts


def get_defaults():
    return {
        "paths": [os.getcwd()],
        "config": [],
        "allow_long_titles": False,
        "ignore": [],
        "sphinx": True,
        "ignore_path": [],
        "ignore_path_errors": [],
        "default_extension": "",
        "file_encoding": "",
        "max_line_length": MAX_LINE_LENGTH,
        "extension": list(FILE_PATTERNS),
        "quiet": False,
        "verbose": False,
        "version": False,
    }


class Result(object):
    def __init__(self):
        self.files_selected = 0
        self.files_ignored = 0
        self.error_counts = {}
        self.errors = []
        self.capture = False

    @property
    def total_errors(self):
        return len(self.errors)

    def error(self, check_name, filename, line_num, code, message):
        self.errors.append((check_name, filename, line_num, code, message))

    def finish(self, files_selected, files_ignored, error_counts):
        self.files_selected = files_selected
        self.files_ignored = files_ignored
        self.error_counts = error_counts

    def report(self):
        lines = []
        if self.capture:
            for error in self.errors:
                lines.append("%s:%s: %s %s" % error[1:])

        lines.extend(
            [
                "=" * 8,
                "Total files scanned = %s" % (self.files_selected),
                "Total files ignored = %s" % (self.files_ignored),
                "Total accumulated errors = %s" % (self.total_errors),
            ]
        )

        if self.error_counts:
            lines.append("Detailed error counts:")
            for check_name in sorted(self.error_counts.keys()):
                check_errors = self.error_counts[check_name]
                lines.append("    - %s = %s" % (check_name, check_errors))

        return "\n".join(lines)


def doc8(args=None, **kwargs):
    result = Result()
    if args is None:
        args = get_defaults()

        # Force reporting to suppress all output
        kwargs["quiet"] = True
        kwargs["verbose"] = False
        result.capture = True

    args["ignore"] = merge_sets(args["ignore"])
    cfg = extract_config(args)
    args["ignore"].update(cfg.pop("ignore", set()))
    if "sphinx" in cfg:
        args["sphinx"] = cfg.pop("sphinx")
    args["extension"].extend(cfg.pop("extension", []))
    args["ignore_path"].extend(cfg.pop("ignore_path", []))

    cfg.setdefault("ignore_path_errors", {})
    tmp_ignores = parse_ignore_path_errors(args.pop("ignore_path_errors", []))
    for path, ignores in tmp_ignores.items():
        if path in cfg["ignore_path_errors"]:
            cfg["ignore_path_errors"][path].update(ignores)
        else:
            cfg["ignore_path_errors"][path] = set(ignores)

    args.update(cfg)

    # Override args with any kwargs
    args.update(kwargs.items())
    setup_logging(args.get("verbose"))

    files, files_ignored = scan(args)
    files_selected = len(files)

    error_counts = validate(args, files, result=result)
    result.finish(files_selected, files_ignored, error_counts)
    return result


def main():
    defaults = get_defaults()
    parser = argparse.ArgumentParser(
        prog="doc8",
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "paths",
        metavar="path",
        type=str,
        nargs="*",
        help=("path to scan for doc files (default: current directory)."),
        default=defaults["paths"],
    )
    parser.add_argument(
        "--config",
        metavar="path",
        action="append",
        help="user config file location"
        " (default: %s)." % ", ".join(CONFIG_FILENAMES),
        default=defaults["config"],
    )
    parser.add_argument(
        "--allow-long-titles",
        action="store_true",
        help="allow long section titles (default: false).",
        default=defaults["allow_long_titles"],
    )
    parser.add_argument(
        "--ignore",
        action="append",
        metavar="code",
        help="ignore the given error code(s).",
        type=split_set_type,
        default=defaults["ignore"],
    )
    parser.add_argument(
        "--no-sphinx",
        action="store_false",
        help="do not ignore sphinx specific false positives.",
        default=defaults["sphinx"],
        dest="sphinx",
    )
    parser.add_argument(
        "--ignore-path",
        action="append",
        default=defaults["ignore_path"],
        help="ignore the given directory or file (globs are supported).",
        metavar="path",
    )
    parser.add_argument(
        "--ignore-path-errors",
        action="append",
        default=defaults["ignore_path_errors"],
        help="ignore the given specific errors in the provided file.",
        metavar="path",
    )
    parser.add_argument(
        "--default-extension",
        action="store",
        help="default file extension to use when a file is"
        " found without a file extension.",
        default=defaults["default_extension"],
        dest="default_extension",
        metavar="extension",
    )
    parser.add_argument(
        "--file-encoding",
        action="store",
        help="set input files text encoding",
        default=defaults["file_encoding"],
        dest="file_encoding",
        metavar="encoding",
    )
    parser.add_argument(
        "--max-line-length",
        action="store",
        metavar="int",
        type=int,
        help="maximum allowed line"
        " length (default: %s)." % defaults["max_line_length"],
        default=defaults["max_line_length"],
    )
    parser.add_argument(
        "-e",
        "--extension",
        action="append",
        metavar="extension",
        help="check file extensions of the given type"
        " (default: %s)." % ", ".join(defaults["extension"]),
        default=defaults["extension"],
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="only print violations",
        default=defaults["quiet"],
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="run in verbose mode.",
        default=defaults["verbose"],
    )
    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="show the version and exit.",
        default=defaults["version"],
    )
    args = vars(parser.parse_args())
    if args.get("version"):
        print(version.version_string)
        return 0

    result = doc8(args)

    if not args.get("quiet"):
        print(result.report())

    if result.total_errors:
        return 1
    else:
        return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

#    Copyright (C) 2014 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

import testtools

from doc8 import checks
from doc8 import parser


class TestTrailingWhitespace(testtools.TestCase):
    def test_trailing(self):
        lines = ["a b  ", "ab"]
        check = checks.CheckTrailingWhitespace({})
        errors = []
        for line in lines:
            errors.extend(check.report_iter(line))
        self.assertEqual(1, len(errors))
        (code, msg) = errors[0]
        self.assertIn(code, check.REPORTS)


class TestTabIndentation(testtools.TestCase):
    def test_tabs(self):
        lines = ["    b", "\tabc", "efg", "\t\tc"]
        check = checks.CheckIndentationNoTab({})
        errors = []
        for line in lines:
            errors.extend(check.report_iter(line))
        self.assertEqual(2, len(errors))
        (code, msg) = errors[0]
        self.assertIn(code, check.REPORTS)


class TestCarriageReturn(testtools.TestCase):
    def test_cr(self):
        content = b"Windows line ending\r\nLegacy Mac line ending\r"
        content += (b"a" * 79) + b"\r\n" + b"\r"
        conf = {"max_line_length": 79}
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(content)
            fh.flush()
            parsed_file = parser.ParsedFile(fh.name)
            check = checks.CheckCarriageReturn(conf)
            errors = list(check.report_iter(parsed_file))
            self.assertEqual(4, len(errors))
            (line, code, msg) = errors[0]
            self.assertIn(code, check.REPORTS)


class TestLineLength(testtools.TestCase):
    def test_over_length(self):
        content = b"""
===
aaa
===

----
test
----

"""
        content += b"\n\n"
        content += (b"a" * 60) + b" " + (b"b" * 60)
        content += b"\n"
        conf = {"max_line_length": 79, "allow_long_titles": True}
        for ext in [".rst", ".txt"]:
            with tempfile.NamedTemporaryFile(suffix=ext) as fh:
                fh.write(content)
                fh.flush()

                parsed_file = parser.ParsedFile(fh.name)
                check = checks.CheckMaxLineLength(conf)
                errors = list(check.report_iter(parsed_file))
                self.assertEqual(1, len(errors))
                (line, code, msg) = errors[0]
                self.assertIn(code, check.REPORTS)

    def test_correct_length(self):
        conf = {"max_line_length": 79, "allow_long_titles": True}
        with tempfile.NamedTemporaryFile(suffix=".rst") as fh:
            fh.write(
                b"known exploit in the wild, for example"
                b" \xe2\x80\x93 the time"
                b" between advance notification"
            )
            fh.flush()

            parsed_file = parser.ParsedFile(fh.name, encoding="utf-8")
            check = checks.CheckMaxLineLength(conf)
            errors = list(check.report_iter(parsed_file))
            self.assertEqual(0, len(errors))

    def test_ignore_code_block(self):
        conf = {"max_line_length": 79, "allow_long_titles": True}
        with tempfile.NamedTemporaryFile(suffix=".rst") as fh:
            fh.write(
                b"List which contains items with code-block\n"
                b"- this is a list item\n\n"
                b"  .. code-block:: ini\n\n"
                b"     this line exceeds 80 chars but should be ignored"
                b"this line exceeds 80 chars but should be ignored"
                b"this line exceeds 80 chars but should be ignored"
            )
            fh.flush()

            parsed_file = parser.ParsedFile(fh.name, encoding="utf-8")
            check = checks.CheckMaxLineLength(conf)
            errors = list(check.report_iter(parsed_file))
            self.assertEqual(0, len(errors))

    def test_unsplittable_length(self):
        content = b"""
===
aaa
===

----
test
----

"""
        content += b"\n\n"
        content += b"a" * 100
        content += b"\n"
        conf = {"max_line_length": 79, "allow_long_titles": True}
        # This number is different since rst parsing is aware that titles
        # are allowed to be over-length, while txt parsing is not aware of
        # this fact (since it has no concept of title sections).
        extensions = [(0, ".rst"), (1, ".txt")]
        for expected_errors, ext in extensions:
            with tempfile.NamedTemporaryFile(suffix=ext) as fh:
                fh.write(content)
                fh.flush()

                parsed_file = parser.ParsedFile(fh.name)
                check = checks.CheckMaxLineLength(conf)
                errors = list(check.report_iter(parsed_file))
                self.assertEqual(expected_errors, len(errors))

    def test_definition_term_length(self):
        conf = {"max_line_length": 79, "allow_long_titles": True}
        with tempfile.NamedTemporaryFile(suffix=".rst") as fh:
            fh.write(
                b"Definition List which contains long term.\n\n"
                b"looooooooooooooooooooooooooooooong definition term"
                b"this line exceeds 80 chars but should be ignored\n"
                b" this is a definition\n"
            )
            fh.flush()

            parsed_file = parser.ParsedFile(fh.name, encoding="utf-8")
            check = checks.CheckMaxLineLength(conf)
            errors = list(check.report_iter(parsed_file))
            self.assertEqual(0, len(errors))


class TestNewlineEndOfFile(testtools.TestCase):
    def test_newline(self):
        tests = [
            (1, b"testing"),
            (1, b"testing\ntesting"),
            (0, b"testing\n"),
            (0, b"testing\ntesting\n"),
            (0, b"testing\r\n"),
            (0, b"testing\r"),
        ]

        for expected_errors, line in tests:
            with tempfile.NamedTemporaryFile() as fh:
                fh.write(line)
                fh.flush()
                parsed_file = parser.ParsedFile(fh.name)
                check = checks.CheckNewlineEndOfFile({})
                errors = list(check.report_iter(parsed_file))
                self.assertEqual(expected_errors, len(errors))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ["sphinx.ext.autodoc"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"doc8"
copyright = u"2013, OpenStack Foundation"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "sphinx_rtd_theme"



# Copyright (C) 2014 Ivan Melnikov <iv at altlinux dot org>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import collections
import re

from docutils import nodes as docutils_nodes

from doc8 import utils


class ContentCheck(metaclass=abc.ABCMeta):
    def __init__(self, cfg):
        self._cfg = cfg

    @abc.abstractmethod
    def report_iter(self, parsed_file):
        pass


class LineCheck(metaclass=abc.ABCMeta):
    def __init__(self, cfg):
        self._cfg = cfg

    @abc.abstractmethod
    def report_iter(self, line):
        pass


class CheckTrailingWhitespace(LineCheck):
    _TRAILING_WHITESPACE_REGEX = re.compile(r"\s$")
    REPORTS = frozenset(["D002"])

    def report_iter(self, line):
        if self._TRAILING_WHITESPACE_REGEX.search(line):
            yield ("D002", "Trailing whitespace")


class CheckIndentationNoTab(LineCheck):
    _STARTING_WHITESPACE_REGEX = re.compile(r"^(\s+)")
    REPORTS = frozenset(["D003"])

    def report_iter(self, line):
        match = self._STARTING_WHITESPACE_REGEX.search(line)
        if match:
            spaces = match.group(1)
            if "\t" in spaces:
                yield ("D003", "Tabulation used for indentation")


class CheckCarriageReturn(ContentCheck):
    REPORTS = frozenset(["D004"])

    def report_iter(self, parsed_file):
        for i, line in enumerate(parsed_file.lines):
            if b"\r" in line:
                yield (i + 1, "D004", "Found literal carriage return")


class CheckNewlineEndOfFile(ContentCheck):
    REPORTS = frozenset(["D005"])

    def __init__(self, cfg):
        super(CheckNewlineEndOfFile, self).__init__(cfg)

    def report_iter(self, parsed_file):
        if parsed_file.lines and not (
            parsed_file.lines[-1].endswith(b"\n")
            or parsed_file._lines[-1].endswith(b"\r")
        ):
            yield (len(parsed_file.lines), "D005", "No newline at end of file")


class CheckValidity(ContentCheck):
    REPORTS = frozenset(["D000"])
    EXT_MATCHER = re.compile(r"(.*)[.]rst", re.I)

    # From docutils docs:
    #
    # Report system messages at or higher than <level>: "info" or "1",
    # "warning"/"2" (default), "error"/"3", "severe"/"4", "none"/"5"
    #
    # See: http://docutils.sourceforge.net/docs/user/config.html#report-level
    WARN_LEVELS = frozenset([2, 3, 4])

    # Only used when running in sphinx mode.
    SPHINX_IGNORES_REGEX = [
        re.compile(r"^Unknown interpreted text"),
        re.compile(r"^Unknown directive type"),
        re.compile(r"^Undefined substitution"),
        re.compile(r"^Substitution definition contains illegal element"),
        re.compile(
            r'^Error in \"code-block\" directive\:\nunknown option: "caption".',
            re.MULTILINE,
        ),
        re.compile(
            r'^Error in "code-block" directive:\nunknown option: "emphasize-lines"'
        ),
        re.compile(r'^Error in "code-block" directive:\nunknown option: "linenos"'),
        re.compile(
            r'^Error in "code-block" directive:\nunknown option: "lineno-start"'
        ),
        re.compile(r'^Error in "code-block" directive:\nunknown option: "dedent"'),
        re.compile(r'^Error in "code-block" directive:\nunknown option: "force"'),
        re.compile(r'^Error in "math" directive:\nunknown option: "label"'),
        re.compile(r'^Error in "math" directive:\nunknown option: "nowrap"'),
        re.compile(
            r'^Error in \"code-block\" directive\:\nunknown option: "substitutions".',
            re.MULTILINE,
        ),
    ]

    def __init__(self, cfg):
        super(CheckValidity, self).__init__(cfg)
        self._sphinx_mode = cfg.get("sphinx")

    def report_iter(self, parsed_file):
        for error in parsed_file.errors:
            if error.level not in self.WARN_LEVELS:
                continue
            ignore = False
            if self._sphinx_mode:
                for m in self.SPHINX_IGNORES_REGEX:
                    if m.match(error.message):
                        ignore = True
                        break
            if not ignore:
                yield (error.line, "D000", error.message)


class CheckMaxLineLength(ContentCheck):
    REPORTS = frozenset(["D001"])

    def __init__(self, cfg):
        super(CheckMaxLineLength, self).__init__(cfg)
        self._max_line_length = self._cfg["max_line_length"]
        self._allow_long_titles = self._cfg["allow_long_titles"]

    def _extract_node_lines(self, doc):
        def extract_lines(node, start_line):
            lines = [start_line]
            if isinstance(node, (docutils_nodes.title)):
                start = start_line - len(node.rawsource.splitlines())
                if start >= 0:
                    lines.append(start)
            if isinstance(node, (docutils_nodes.literal_block)):
                end = start_line + len(node.rawsource.splitlines()) - 1
                lines.append(end)
            return lines

        def gather_lines(node):
            lines = []
            for n in node.traverse(include_self=True):
                lines.extend(extract_lines(n, find_line(n)))
            return lines

        def find_line(node):
            n = node
            while n is not None:
                if n.line is not None:
                    return n.line
                n = n.parent
            return None

        def filter_systems(node):
            if utils.has_any_node_type(node, (docutils_nodes.system_message,)):
                return False
            return True

        nodes_lines = []
        first_line = -1
        for n in utils.filtered_traverse(doc, filter_systems):
            line = find_line(n)
            if line is None:
                continue
            if first_line == -1:
                first_line = line
            contained_lines = set(gather_lines(n))
            nodes_lines.append((n, (min(contained_lines), max(contained_lines))))
        return (nodes_lines, first_line)

    def _extract_directives(self, lines):
        def starting_whitespace(line):
            m = re.match(r"^(\s+)(.*)$", line)
            if not m:
                return 0
            return len(m.group(1))

        def all_whitespace(line):
            return bool(re.match(r"^(\s*)$", line))

        def find_directive_end(start, lines):
            after_lines = collections.deque(lines[start + 1 :])
            k = 0
            while after_lines:
                line = after_lines.popleft()
                if all_whitespace(line) or starting_whitespace(line) >= 1:
                    k += 1
                else:
                    break
            return start + k

        # Find where directives start & end so that we can exclude content in
        # these directive regions (the rst parser may not handle this correctly
        # for unknown directives, so we have to do it manually).
        directives = []
        for i, line in enumerate(lines):
            if re.match(r"^\s*..\s(.*?)::\s*", line):
                directives.append((i, find_directive_end(i, lines)))
            elif re.match(r"^::\s*$", line):
                directives.append((i, find_directive_end(i, lines)))

        # Find definition terms in definition lists
        # This check may match the code, which is already appended
        lwhitespaces = r"^\s*"
        listspattern = r"^\s*(\* |- |#\. |\d+\. )"
        for i in range(0, len(lines) - 1):
            line = lines[i]
            next_line = lines[i + 1]
            # if line is a blank, line is not a definition term
            if all_whitespace(line):
                continue
            # if line is a list, line is checked as normal line
            if re.match(listspattern, line):
                continue
            if len(re.search(lwhitespaces, line).group()) < len(
                re.search(lwhitespaces, next_line).group()
            ):
                directives.append((i, i))

        return directives

    def _txt_checker(self, parsed_file):
        for i, line in enumerate(parsed_file.lines_iter()):
            if len(line) > self._max_line_length:
                if not utils.contains_url(line):
                    yield (i + 1, "D001", "Line too long")

    def _rst_checker(self, parsed_file):
        lines = list(parsed_file.lines_iter())
        doc = parsed_file.document
        nodes_lines, first_line = self._extract_node_lines(doc)
        directives = self._extract_directives(lines)

        def find_containing_nodes(num):
            if num < first_line and len(nodes_lines):
                return [nodes_lines[0][0]]
            contained_in = []
            for (n, (line_min, line_max)) in nodes_lines:
                if num >= line_min and num <= line_max:
                    contained_in.append((n, (line_min, line_max)))
            smallest_span = None
            best_nodes = []
            for (n, (line_min, line_max)) in contained_in:
                span = line_max - line_min
                if smallest_span is None:
                    smallest_span = span
                    best_nodes = [n]
                elif span < smallest_span:
                    smallest_span = span
                    best_nodes = [n]
                elif span == smallest_span:
                    best_nodes.append(n)
            return best_nodes

        def any_types(nodes, types):
            return any([isinstance(n, types) for n in nodes])

        skip_types = (docutils_nodes.target, docutils_nodes.literal_block)
        title_types = (
            docutils_nodes.title,
            docutils_nodes.subtitle,
            docutils_nodes.section,
        )
        for i, line in enumerate(lines):
            if len(line) > self._max_line_length:
                in_directive = False
                for (start, end) in directives:
                    if i >= start and i <= end:
                        in_directive = True
                        break
                if in_directive:
                    continue
                stripped = line.lstrip()
                if " " not in stripped:
                    # No room to split even if we could.
                    continue
                if utils.contains_url(stripped):
                    continue
                nodes = find_containing_nodes(i + 1)
                if any_types(nodes, skip_types):
                    continue
                if self._allow_long_titles and any_types(nodes, title_types):
                    continue
                yield (i + 1, "D001", "Line too long")

    def report_iter(self, parsed_file):
        if parsed_file.extension.lower() != ".rst":
            checker_func = self._txt_checker
        else:
            checker_func = self._rst_checker
        for issue in checker_func(parsed_file):
            yield issue



# -*- coding: utf-8 -*-

from mock import patch, MagicMock
import os
from io import StringIO
import shutil
import sys
import testtools

from doc8.main import main, doc8


# Location to create test files
TMPFS_DIR_NAME = ".tmp"

# Expected output
OUTPUT_CMD_NO_QUIET = """\
Scanning...
Validating...
{path}/invalid.rst:1: D002 Trailing whitespace
{path}/invalid.rst:1: D005 No newline at end of file
========
Total files scanned = 1
Total files ignored = 0
Total accumulated errors = 2
Detailed error counts:
    - doc8.checks.CheckCarriageReturn = 0
    - doc8.checks.CheckIndentationNoTab = 0
    - doc8.checks.CheckMaxLineLength = 0
    - doc8.checks.CheckNewlineEndOfFile = 1
    - doc8.checks.CheckTrailingWhitespace = 1
    - doc8.checks.CheckValidity = 0
"""

OUTPUT_CMD_QUIET = """\
{path}/invalid.rst:1: D002 Trailing whitespace
{path}/invalid.rst:1: D005 No newline at end of file
"""

OUTPUT_CMD_VERBOSE = """\
Scanning...
  Selecting '{path}/invalid.rst'
Validating...
Validating {path}/invalid.rst (utf-8, 10 chars, 1 lines)
  Running check 'doc8.checks.CheckValidity'
  Running check 'doc8.checks.CheckTrailingWhitespace'
    - {path}/invalid.rst:1: D002 Trailing whitespace
  Running check 'doc8.checks.CheckIndentationNoTab'
  Running check 'doc8.checks.CheckCarriageReturn'
  Running check 'doc8.checks.CheckMaxLineLength'
  Running check 'doc8.checks.CheckNewlineEndOfFile'
    - {path}/invalid.rst:1: D005 No newline at end of file
========
Total files scanned = 1
Total files ignored = 0
Total accumulated errors = 2
Detailed error counts:
    - doc8.checks.CheckCarriageReturn = 0
    - doc8.checks.CheckIndentationNoTab = 0
    - doc8.checks.CheckMaxLineLength = 0
    - doc8.checks.CheckNewlineEndOfFile = 1
    - doc8.checks.CheckTrailingWhitespace = 1
    - doc8.checks.CheckValidity = 0
"""

OUTPUT_API_REPORT = """\
{path}/invalid.rst:1: D002 Trailing whitespace
{path}/invalid.rst:1: D005 No newline at end of file
========
Total files scanned = 1
Total files ignored = 0
Total accumulated errors = 2
Detailed error counts:
    - doc8.checks.CheckCarriageReturn = 0
    - doc8.checks.CheckIndentationNoTab = 0
    - doc8.checks.CheckMaxLineLength = 0
    - doc8.checks.CheckNewlineEndOfFile = 1
    - doc8.checks.CheckTrailingWhitespace = 1
    - doc8.checks.CheckValidity = 0"""


class Capture(object):
    """
    Context manager to capture output on stdout and stderr
    """

    def __enter__(self):
        self.old_out = sys.stdout
        self.old_err = sys.stderr
        self.out = StringIO()
        self.err = StringIO()

        sys.stdout = self.out
        sys.stderr = self.err

        return self.out, self.err

    def __exit__(self, *args):
        sys.stdout = self.out
        sys.stderr = self.err


class TmpFs(object):
    """
    Context manager to create and clean a temporary file area for testing
    """

    def __enter__(self):
        self.path = os.path.join(os.getcwd(), TMPFS_DIR_NAME)
        if os.path.exists(self.path):
            raise ValueError(
                "Tmp dir found at %s - remove before running tests" % self.path
            )
        os.mkdir(self.path)
        return self

    def __exit__(self, *args):
        shutil.rmtree(self.path)

    def create_file(self, filename, content):
        with open(os.path.join(self.path, filename), "w") as file:
            file.write(content)

    def mock(self):
        """
        Create a file which fails on a LineCheck and a ContentCheck
        """
        self.create_file("invalid.rst", "D002 D005 ")

    def expected(self, template):
        """
        Insert the path into a template to generate an expected test value
        """
        return template.format(path=self.path)


class FakeResult(object):
    """
    Minimum valid result returned from doc8
    """

    total_errors = 0

    def report(self):
        return ""


class TestCommandLine(testtools.TestCase):
    """
    Test command line invocation
    """

    def test_main__no_quiet_no_verbose__output_is_not_quiet(self):
        with TmpFs() as tmpfs, Capture() as (out, err), patch(
            "argparse._sys.argv", ["doc8", tmpfs.path]
        ):
            tmpfs.mock()
            state = main()
            self.assertEqual(out.getvalue(), tmpfs.expected(OUTPUT_CMD_NO_QUIET))
            self.assertEqual(err.getvalue(), "")
        self.assertEqual(state, 1)

    def test_main__quiet_no_verbose__output_is_quiet(self):
        with TmpFs() as tmpfs, Capture() as (out, err), patch(
            "argparse._sys.argv", ["doc8", "--quiet", tmpfs.path]
        ):
            tmpfs.mock()
            state = main()
            self.assertEqual(out.getvalue(), tmpfs.expected(OUTPUT_CMD_QUIET))
            self.assertEqual(err.getvalue(), "")
        self.assertEqual(state, 1)

    def test_main__no_quiet_verbose__output_is_verbose(self):
        with TmpFs() as tmpfs, Capture() as (out, err), patch(
            "argparse._sys.argv", ["doc8", "--verbose", tmpfs.path]
        ):
            tmpfs.mock()
            state = main()
            self.assertEqual(out.getvalue(), tmpfs.expected(OUTPUT_CMD_VERBOSE))
            self.assertEqual(err.getvalue(), "")
        self.assertEqual(state, 1)


class TestApi(testtools.TestCase):
    """
    Test direct code invocation
    """

    def test_doc8__defaults__no_output_and_report_as_expected(self):
        with TmpFs() as tmpfs, Capture() as (out, err):
            tmpfs.mock()
            result = doc8(paths=[tmpfs.path])

            self.assertEqual(result.report(), tmpfs.expected(OUTPUT_API_REPORT))
            self.assertEqual(
                result.errors,
                [
                    (
                        "doc8.checks.CheckTrailingWhitespace",
                        "{}/invalid.rst".format(tmpfs.path),
                        1,
                        "D002",
                        "Trailing whitespace",
                    ),
                    (
                        "doc8.checks.CheckNewlineEndOfFile",
                        "{}/invalid.rst".format(tmpfs.path),
                        1,
                        "D005",
                        "No newline at end of file",
                    ),
                ],
            )
            self.assertEqual(out.getvalue(), "")
            self.assertEqual(err.getvalue(), "")
        self.assertEqual(result.total_errors, 2)

    def test_doc8__verbose__verbose_overridden(self):
        with TmpFs() as tmpfs, Capture() as (out, err):
            tmpfs.mock()
            result = doc8(paths=[tmpfs.path], verbose=True)

            self.assertEqual(result.report(), tmpfs.expected(OUTPUT_API_REPORT))
            self.assertEqual(out.getvalue(), "")
            self.assertEqual(err.getvalue(), "")
        self.assertEqual(result.total_errors, 2)


class TestArguments(testtools.TestCase):
    """
    Test that arguments are parsed correctly
    """

    def get_args(self, **kwargs):
        # Defaults
        args = {
            "paths": [os.getcwd()],
            "config": [],
            "allow_long_titles": False,
            "ignore": set(),
            "sphinx": True,
            "ignore_path": [],
            "ignore_path_errors": {},
            "default_extension": "",
            "file_encoding": "",
            "max_line_length": 79,
            "extension": list([".rst", ".txt"]),
            "quiet": False,
            "verbose": False,
            "version": False,
        }
        args.update(kwargs)
        return args

    def test_get_args__override__value_overridden(self):
        # Just checking a dict is a dict, but confirms nothing silly has happened
        # so we can make assumptions about get_args in other tests
        self.assertEqual(
            self.get_args(paths=["/tmp/does/not/exist"]),
            {
                "paths": ["/tmp/does/not/exist"],
                "config": [],
                "allow_long_titles": False,
                "ignore": set(),
                "sphinx": True,
                "ignore_path": [],
                "ignore_path_errors": {},
                "default_extension": "",
                "file_encoding": "",
                "max_line_length": 79,
                "extension": [".rst", ".txt"],
                "quiet": False,
                "verbose": False,
                "version": False,
            },
        )

    def test_args__no_args__defaults(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch("argparse._sys.argv", ["doc8"]):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args())

    def test_args__paths__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "path1", "path2"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(paths=["path1", "path2"]))

    def test_args__config__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        mock_config = MagicMock(return_value={})
        with patch("doc8.main.scan", mock_scan), patch(
            "doc8.main.extract_config", mock_config
        ), patch(
            "argparse._sys.argv", ["doc8", "--config", "path1", "--config", "path2"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(config=["path1", "path2"]))

    def test_args__allow_long_titles__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--allow-long-titles"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(allow_long_titles=True))

    def test_args__ignore__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv",
            ["doc8", "--ignore", "D002", "--ignore", "D002", "--ignore", "D005"],
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(ignore={"D002", "D005"}))

    def test_args__sphinx__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--no-sphinx"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(sphinx=False))

    def test_args__ignore_path__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv",
            ["doc8", "--ignore-path", "path1", "--ignore-path", "path2"],
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(
                self.get_args(ignore_path=["path1", "path2"])
            )

    def test_args__ignore_path_errors__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv",
            [
                "doc8",
                "--ignore-path-errors",
                "path1;D002",
                "--ignore-path-errors",
                "path2;D005",
            ],
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(
                self.get_args(ignore_path_errors={"path1": {"D002"}, "path2": {"D005"}})
            )

    def test_args__default_extension__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--default-extension", "rst"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(default_extension="rst"))

    def test_args__file_encoding__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--file-encoding", "utf8"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(file_encoding="utf8"))

    def test_args__max_line_length__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--max-line-length", "88"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(max_line_length=88))

    def test_args__extension__overrides_default(self):
        # ": [".rst", ".txt"],
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--extension", "ext1", "--extension", "ext2"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(
                self.get_args(extension=[".rst", ".txt", "ext1", "ext2"])
            )

    def test_args__quiet__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--quiet"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(quiet=True))

    def test_args__verbose__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--verbose"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_called_once_with(self.get_args(verbose=True))

    def test_args__version__overrides_default(self):
        mock_scan = MagicMock(return_value=([], 0))
        with patch("doc8.main.scan", mock_scan), patch(
            "argparse._sys.argv", ["doc8", "--version"]
        ):
            state = main()
            self.assertEqual(state, 0)
            mock_scan.assert_not_called()
